# DO NOT EDIT
#
# GENERATED BY: Inline::Module 0.21
#
# This module is for author-side development only. When this module is shipped
# to CPAN, it will be automagically replaced with content that does not
# require any Inline framework modules (or any other non-core modules).

use strict; use warnings;
package Acme::Math::XS::Inline;
use Inline::Module 'v1' => '0.21';

1;
